import json
import logging

import numpy as np
import treefiles as tf

from gcma.io.BaseIO import IOEncoder
from gcma.io.Feature import TFeatures, Features
from gcma.io.Param import TParams, Params


class Hist:
    def __init__(
        self, params: TParams, features: TFeatures, fvalue: float, it: int = None
    ):
        self.params = params
        self.features = features
        self.fvalue = fvalue
        self.iteration = it

    def to_dict(self) -> dict:
        return {
            "iteration": self.iteration,
            "params": self.params,
            "features": self.features,
            "fvalue": self.fvalue,
        }

    @classmethod
    def from_dict(cls, d: dict):
        return cls(
            Params(d["params"]),
            Features(d["features"]),
            float(d["fvalue"]),
            it=int(d["iteration"]),
        )

    def __repr__(self):
        s = f"Params:\n{Params(self.params).table()}\n"
        s += f"Features:\n{Features(self.features).table()}"
        return s


class HistoryEncoder(IOEncoder):
    def default(self, obj):
        if isinstance(obj, Hist):
            return obj.to_dict()
        return super().default(obj)


class History(list):
    @property
    def best(self):
        """
        Return the best history among all generations
        """
        fvalues = np.array([x.fvalue for x in self])
        return self[np.argmin(fvalues)]

    @property
    def last(self):
        """
        Return the best history of last generation
        """
        return self[-1]

    def add(self, params: TParams, features: TFeatures, fvalue: float):
        self.append(Hist(params, features, fvalue, it=len(self)))

    def dump(self, fname: str):
        tf.dump_json(fname, self, cls=HistoryEncoder)

    @classmethod
    def load(cls, fname: str):
        l = [Hist.from_dict(x) for x in tf.load_json(fname)]
        return History(l)


log = logging.getLogger(__name__)
